# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-08-21 13:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('identities', '0003_optout'),
    ]

    operations = [
        migrations.CreateModel(
            name='DetailKey',
            fields=[
                ('key_name', models.CharField(max_length=200, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]
